from hierarchybuilder.topic_clustering import main_clustering as main_clustering
from hierarchybuilder.expansions import parse_medical_data
import torch
import os
import sys
from transformers import AutoTokenizer, AutoModel
import statistics

cos = torch.nn.CosineSimilarity(dim=0, eps=1e-08)
nlp = parse_medical_data.nlp
device = torch.device("cpu")
# device = torch.device("cuda:1" if torch.cuda.is_available() else "cpu")
sapBert_tokenizer = AutoTokenizer.from_pretrained('cambridgeltl/SapBERT-from-PubMedBERT-fulltext')
sapBert_model = AutoModel.from_pretrained('cambridgeltl/SapBERT-from-PubMedBERT-fulltext')
model = sapBert_model
model = model.eval()

dict_span_to_lemma_lst = {}
topics_dict = {}
dict_span_to_counter = {}
dict_word_to_lemma = {}
dict_lemma_to_synonyms = {}
dict_longest_span_to_counter = {}
dict_noun_lemma_to_synonyms = {}
dict_noun_lemma_to_noun_words = {}
dict_noun_lemma_to_counter = {}
dict_noun_word_to_counter = {}
dict_full_np_to_sentences = {}
entries_number_limit = 50
host_and_port = "127.0.0.1:5000"


def initialize_data(examples, host_val, port_val, ignore_words, entries_number,
                    device_type, has_umls_server):
    global topics_dict, dict_span_to_counter, dict_word_to_lemma, dict_lemma_to_synonyms, \
        dict_longest_span_to_counter, dict_noun_lemma_to_synonyms, dict_noun_lemma_to_noun_words, \
        dict_noun_lemma_to_counter, dict_noun_word_to_counter, entries_number_limit, device, model, \
        host_and_port, dict_full_np_to_sentences
    host_and_port = "http://" + host_val + ":" + str(port_val)
    if device_type:
        device = device_type
        model = model.to(device)
        model = model.eval()
    entries_number_limit = entries_number
    if ignore_words is None:
        ignore_words = set()
    collection_format_examples = parse_medical_data.get_examples_as_all_optional_answers_format(examples)
    topics_dict, dict_span_to_counter, dict_word_to_lemma, dict_lemma_to_synonyms, \
    dict_longest_span_to_counter, dict_noun_lemma_to_synonyms, dict_noun_lemma_to_noun_words, dict_noun_lemma_to_counter, \
    dict_noun_word_to_counter, dict_full_np_to_sentences = \
        main_clustering.convert_examples_to_clustered_data(collection_format_examples, ignore_words, host_and_port,
                                                           has_umls_server)
    dict_span_to_counter.update(dict_noun_word_to_counter)
    dict_span_to_counter.update(dict_noun_lemma_to_counter)


def dfs_for_cyclic(visited, helper, node):
    visited.append(node)
    helper.append(node)
    children = node.children
    for child in children:
        if child not in visited:
            ans = dfs_for_cyclic(visited, helper, child)
            if ans == True:
                print(child.span_lst)
                return True
        elif child in helper:
            print(child.span_lst)
            return True
    helper.remove(node)
    return False


def isCyclic(nodes_lst):
    visited = []
    helper = []
    for i in nodes_lst:
        if i not in visited:
            ans = dfs_for_cyclic(visited, helper, i)
            if ans == True:
                print(i.span_lst)
                return True
    return False


def update_nodes_labels(nodes_lst, visited=set()):
    labels_lst = set()
    for node in nodes_lst:
        if node in visited:
            continue
        visited.add(node)
        desc_labels = update_nodes_labels(node.children, visited)
        node.label_lst.update(desc_labels)
        labels_lst.update(node.label_lst)
    return labels_lst


def get_all_spans(np_object_lst, all_spans, visited=set()):
    for np_object in np_object_lst:
        if np_object in visited:
            continue
        visited.add(np_object)
        all_spans.update(np_object.span_lst)
        get_all_spans(np_object.children, all_spans, visited)


def dfs(visited, node):
    if node not in visited:
        visited.append(node)
        for neighbour in node.children:
            dfs(visited, neighbour)


def depth_dag(node, counter=0, visited=set()):
    max_depth = counter
    for child in node.children:
        current_max_depth = depth_dag(child, counter + 1, visited)
        if current_max_depth > max_depth:
            max_depth = current_max_depth
    return max_depth


def get_leaves(node, leaves, visited=set()):
    if node in visited:
        return
    visited.add(node)
    if not node.children:
        leaves.add(node)
        return
    for child in node.children:
        get_leaves(child, leaves, visited)


def get_all_nodes(nodes_lst, visited):
    for node in nodes_lst:
        if node in visited:
            continue
        visited.add(node)
        get_all_nodes(node.children, visited)


def calculation_for_paper(topic_object_lst, top_k_topics):
    visited = set()
    get_all_nodes(topic_object_lst, visited)
    all_dag_nodes = list(visited)
    print("number of nodes is " + str(len(all_dag_nodes)))
    max_depth = 0
    for topic in topic_object_lst:
        depth = depth_dag(topic)
        if depth > max_depth:
            max_depth = depth
    print("the depth of the entire DAG is " + str(max_depth))
    min_leaves = 10000
    max_leaves = 0
    total_leaves = 0
    max_depth = 0
    min_depth = 1000
    total_depth = 0
    all_leaves = set()
    for entry in top_k_topics:
        leaves = set()
        get_leaves(entry, leaves)
        all_leaves.update(leaves)
        leaves_number = len(leaves)
        total_leaves += leaves_number
        if leaves_number < min_leaves:
            min_leaves = leaves_number
        if leaves_number > max_leaves:
            max_leaves = leaves_number
        depth = depth_dag(entry)
        total_depth += depth
        if depth > max_depth:
            max_depth = depth
        if depth < min_depth:
            min_depth = depth
    # top k leaves
    print("average number of leaves is ")
    print(total_leaves / len(top_k_topics))
    print("minimal leaves for top k entry is " + str(min_leaves))
    print("maximal leaves for top k entry is " + str(max_leaves))
    # top k depth
    print("average number of depth is ")
    print(total_depth / len(top_k_topics))
    print("minimal depth for top k entry is " + str(min_depth))
    print("maximal depth for top k entry is " + str(max_depth))
    all_dag_nodes_in_top_k = set()
    get_all_nodes(top_k_topics, all_dag_nodes_in_top_k)
    number_of_children_array = []
    for node in all_dag_nodes_in_top_k:
        if node in top_k_topics:
            continue
        if node in all_leaves:
            continue
        number_of_children_array.append(len(node.children))
    # internal nodes
    print("minimal val of internal nodes is " + str(min(number_of_children_array)))
    print("maximal val of internal nodes is " + str(max(number_of_children_array)))
    print("average number of internal nodes ")
    print(len(number_of_children_array) / len(top_k_topics))
    ans = statistics.variance(number_of_children_array)
    print("The variance of list is : ")
    print(ans)



import requests
import json
from hierarchybuilder.combine_spans import utils as combine_spans_utils
import hierarchybuilder.utils as ut


def combine_nodes_by_umls_synonymous_spans_dfs_helper(topic_object_lst, dict_span_to_object, np_object, visited,
                                                      dict_object_to_global_label,
                                                      global_dict_label_to_object, span_to_vector):
    if np_object in visited:
        return
    visited.add(np_object)
    equivalent_object_lst = set()
    post_data = json.dumps(list(np_object.span_lst))
    dict_response = requests.post(ut.host_and_port + '/create_synonyms_dictionary/',
                                  params={"words": post_data}, timeout=3)
    try:
        output = dict_response.json()
        synonyms_dict = output['synonyms']
    except:
        synonyms_dict = {}
    synonyms = set()
    for key, synonyms_lst in synonyms_dict.items():
        synonyms.update(synonyms_lst)
    if synonyms:
        for term in synonyms:
            equivalent_np_object = dict_span_to_object.get(term, None)
            if equivalent_np_object:
                if equivalent_np_object == np_object:
                    continue
                equivalent_object_lst.add(equivalent_np_object)
        if equivalent_object_lst:
            equivalent_object_lst = [np_object] + list(equivalent_object_lst)
            combined_nodes_lst = set()
            combine_spans_utils.combine_nodes_lst(equivalent_object_lst, dict_span_to_object,
                                                  dict_object_to_global_label,
                                                  global_dict_label_to_object, topic_object_lst, combined_nodes_lst)
            for node in combined_nodes_lst:
                visited.add(node)

    children_lst = np_object.children.copy()
    for child in children_lst:
        combine_nodes_by_umls_synonymous_spans_dfs_helper(topic_object_lst, dict_span_to_object, child, visited,
                                                          dict_object_to_global_label,
                                                          global_dict_label_to_object, span_to_vector)


def combine_nodes_by_umls_synonymous_spans(dict_span_to_object, dict_object_to_global_label,
                                           global_dict_label_to_object,
                                           topic_object_lst, span_to_vector):
    visited = set()
    topic_objects = topic_object_lst.copy()
    for topic_object in topic_objects:
        if topic_object in visited:
            topic_objects.remove(topic_object)
            continue
        combine_nodes_by_umls_synonymous_spans_dfs_helper(topic_object_lst, dict_span_to_object, topic_object, visited,
                                                          dict_object_to_global_label,
                                                          global_dict_label_to_object, span_to_vector)



import itertools

noun_tags_lst = ['NN', 'NNS', 'WP', 'PRP', 'NNP', 'NNPS']


def get_head_of_span(span):
    head = None
    for child in span:
        if child.head == child:
            return child
        if child.head not in span:
            head = child
    return head


def from_children_to_list(children):
    lst_children = []
    for token in children:
        lst_children.append(token)
    return lst_children


def get_token_by_dep(lst_children, dep_type):
    lst_tokens = []
    for child in lst_children:
        if child.dep_ == dep_type:
            lst_tokens.append(child)
    return lst_tokens


def powerset(iterable):
    return itertools.chain.from_iterable(itertools.combinations(iterable, r) for r in range(len(iterable) + 1))


def from_lst_to_sequence(sub_np_lst):
    sub_np_of_child_lst_final = []
    current_lst = []
    if isinstance(sub_np_lst[0], list):
        if len(sub_np_lst) == 1:
            return from_lst_to_sequence(sub_np_lst[0])
        sub_np_of_child_lst = []
        for child in sub_np_lst:
            sub_np_of_child = from_lst_to_sequence(child)
            sub_np_of_child_lst.append(sub_np_of_child)
    else:
        collect_to_lst = []
        slice_index = 0
        for item in sub_np_lst:
            if isinstance(item, list):
                break
            collect_to_lst.append(item)
            slice_index += 1
        current_lst.extend(collect_to_lst)
        if len(sub_np_lst) == 1:
            return [current_lst]
        sub_np_of_child_lst = []
        for child in sub_np_lst[slice_index:]:
            sub_np_of_child = from_lst_to_sequence(child)
            sub_np_of_child_lst.append(sub_np_of_child)
    result_list = list(powerset(sub_np_of_child_lst))
    for item in result_list:
        for element in itertools.product(*item):
            if item == ():
                continue
            lst_temp = []
            for token in element:
                lst_temp.extend(token)
            sub_np_of_child_lst_final.append(lst_temp)
    if current_lst:
        for lst in sub_np_of_child_lst_final:
            lst.extend(current_lst)
        sub_np_of_child_lst_final.append(current_lst)
    return sub_np_of_child_lst_final


def is_np_child_head(head_word, word):
    if word.head == head_word:
        return True
    if word == word.head:
        return False
    return is_np_child_head(head_word, word.head)


def get_np_boundary(head_word_index, sentence_dep_graph):
    head_word = sentence_dep_graph[head_word_index]
    if head_word.tag_ not in noun_tags_lst:
        return None, None, None
    boundary_np_to_the_left = head_word_index
    # np's boundary to the left
    for i in range(1, head_word_index + 1):
        current_index = head_word_index - i
        word = sentence_dep_graph[current_index]
        if is_np_child_head(head_word, word):
            boundary_np_to_the_left = current_index
            continue
        break
    # np's boundary to the right
    boundary_np_to_the_right = head_word_index
    for i in range(head_word_index + 1, len(sentence_dep_graph)):
        word = sentence_dep_graph[i]
        if is_np_child_head(head_word, word):
            boundary_np_to_the_right = i
            continue
        break
    while sentence_dep_graph[boundary_np_to_the_left].dep_ in ['case', 'mark'] or sentence_dep_graph[boundary_np_to_the_left].tag_ == 'IN':
        boundary_np_to_the_left += 1
    return sentence_dep_graph[
           boundary_np_to_the_left:boundary_np_to_the_right + 1], head_word_index - boundary_np_to_the_left, boundary_np_to_the_left


def get_tokens_as_span(tokens):
    span = ""
    idx = 0
    for token in tokens:
        if idx == 0 and token.tag_ in ['IN', 'TO']:
            continue
        if idx != 0 and token.text not in [',', '(', ')', '.', '/'] and not not_space:
            span += ' '
        span += token.text.lower()
        idx += 1
        if token.text in ['(''/']:
            not_space = True
            continue
        not_space = False
    return span


def write_to_file_dict_counter(sub_np_final_lst_collection, output_file):
    with open(output_file, 'w', encoding='utf-8') as f:
        counter = 0
        for sub_np_final_lst in sub_np_final_lst_collection:
            lst_to_check_duplication = []
            for tokens in sub_np_final_lst:
                span = get_tokens_as_span(tokens)
                if span in lst_to_check_duplication:
                    print(span)
                    counter += 1
                else:
                    lst_to_check_duplication.append(span)
                f.write(span)
                f.write('\n')
            f.write('\n')
        print("Num of duplication:", counter)



import hierarchybuilder.DAG.DAG_utils as DAG_utils
import hierarchybuilder.utils as ut
import json
from json import JSONEncoder


def print_flat_list_to_file(concept_to_occurrences):
    concept_to_occurrences = {k: v for k, v in
                              sorted(concept_to_occurrences.items(), key=lambda item: item[1], reverse=True)}
    file_name = "flat_list_for_UI_chest_pain.txt"
    concept_lst = []
    with open(file_name, 'w', encoding='utf-8') as f:
        idx = 0
        for longest_span, number in concept_to_occurrences.items():
            concept = longest_span + ': ' + str(number)
            concept_lst.append(concept)
            concept = str(idx) + ") " + concept
            f.write(concept)
            idx += 1
            f.write('\n')
    with open('flat_list_for_UI_as_json_chest_pain.txt', 'w') as result_file:
        result_file.write(json.dumps(concept_lst))


def get_all_labels(nodes, labels, visited=set()):
    for node in nodes:
        if node in visited:
            continue
        visited.add(node)
        labels.update(node.label_lst)
        get_all_labels(node.children, labels, visited)


class json_top_k_object:
    def __init__(self, json_top_k):
        self.json_top_k = json_top_k

    # def to_json(self):
    #     to_return = self.json_top_k
    #
    #     return to_return
    def to_json(self):
        to_return = {}
        for node in self.json_top_k:
            to_return.update(node.to_json())

        return to_return


def json_dag_visualization(top_k_topics, global_index_to_similar_longest_np, taxonomic_np_objects, topic_object_lst):
    different_concepts = set()
    concept_to_occurrences = {}
    top_k_topics_as_json = DAG_utils.from_DAG_to_JSON(top_k_topics, global_index_to_similar_longest_np,
                                                      taxonomic_np_objects, different_concepts,
                                                      concept_to_occurrences)
    np_object_to_json_node = {}
    DAG_utils.from_DAG_to_json_nested_objects(top_k_topics, global_index_to_similar_longest_np,
                                              np_object_to_json_node)
    json_top_k_nodes = []
    for np_object in top_k_topics:
        json_top_k_nodes.append(np_object_to_json_node[hash(np_object)])
    json_entry_list = json.dumps(json_top_k_object(json_top_k_nodes), default=DAG_utils.default)

    # root = json_top_k_object(json_top_k_nodes)
    top_k_labels = set()
    get_all_labels(top_k_topics, top_k_labels, visited=set())
    covered_labels = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                             top_k_labels)
    labels_of_topics = set()
    get_all_labels(topic_object_lst, labels_of_topics, visited=set())
    total_labels_of_topics = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                                     labels_of_topics)
    print("total labels of topics:", total_labels_of_topics)
    print("Covered labels by selected nodes:", covered_labels)
    print("Done")
    return json_entry_list



from hierarchybuilder.topic_clustering import utils_clustering
from hierarchybuilder.expansions import valid_expansion_utils
from hierarchybuilder.expansions import parse_medical_data


def filter_and_sort_dicts():
    topic_lst = utils_clustering.set_cover()
    utils_clustering.filter_dict_by_lst(topic_lst)


def combine_word_in_upper_case_to_word_in_lower_if_exist(dict_noun_lemma_to_counter, dict_noun_lemma_to_examples,
                                                         upper_case_noun_lst):
    entries_to_remove = set()
    for upper_case_noun_entry in upper_case_noun_lst:
        lower_case_noun_entry = upper_case_noun_entry.lower()
        if lower_case_noun_entry in dict_noun_lemma_to_examples and upper_case_noun_entry in dict_noun_lemma_to_examples:
            entries_to_remove.add(upper_case_noun_entry)
            upper_case_noun_entry_examples = dict_noun_lemma_to_examples[upper_case_noun_entry]
            dict_noun_lemma_to_examples[lower_case_noun_entry].extend(upper_case_noun_entry_examples)
            if lower_case_noun_entry in dict_noun_lemma_to_counter and \
                    upper_case_noun_entry in dict_noun_lemma_to_counter:
                dict_noun_lemma_to_counter[lower_case_noun_entry] += dict_noun_lemma_to_counter[upper_case_noun_entry]
    for entry in entries_to_remove:
        dict_noun_lemma_to_examples.pop(entry, None)
        dict_noun_lemma_to_counter.pop(entry, None)


def create_dictionary_for_abbreviation(examples):
    abbreviation_lst = set()
    noun_words = set()
    dict_compound_noun_to_lst = {}
    for biggest_noun_phrase, head_span, all_valid_nps_lst, sentence in examples:
        compound_noun = utils_clustering.combine_tied_deps_recursively_and_combine_their_children(head_span)
        compound_noun.sort(key=lambda x: x.i)
        compound_as_lemma_lst = []
        for token in compound_noun:
            compound_as_lemma_lst.append(token.text.lower())
        compound_noun_as_span = valid_expansion_utils.get_tokens_as_span(compound_noun)
        dict_compound_noun_to_lst[compound_noun_as_span] = compound_as_lemma_lst
        for token in biggest_noun_phrase:
            if token.pos_ in "NOUN":
                compound_noun = utils_clustering.combine_tied_deps_recursively_and_combine_their_children(token)
                for token in compound_noun:
                    if token.text.isupper():
                        abbreviation_lst.add(token.text)
                        continue
                    noun_words.add(token.lemma_.lower())
    remove_lst = set()
    dict_abbreviation_to_elaboration = {}
    for abbreviation in abbreviation_lst:
        if abbreviation.lower() in noun_words:
            dict_abbreviation_to_elaboration[abbreviation] = abbreviation.lower()
            remove_lst.add(abbreviation)
    abbreviation_lst = abbreviation_lst - remove_lst
    return abbreviation_lst, dict_compound_noun_to_lst, dict_abbreviation_to_elaboration


def replace_abbreviation_with_elaboration(dict_abbreviation_to_elaboration, dict_noun_lemma_to_examples,
                                          dict_elaboration_to_lst, dict_noun_word_to_counter):
    for abbreviation, elaboration in dict_abbreviation_to_elaboration.items():
        for word in elaboration:
            example_lst = dict_noun_lemma_to_examples.get(word, [])
            if not example_lst:
                continue
            example_lst_abbreviation = dict_noun_lemma_to_examples[abbreviation]
            dict_noun_word_to_counter[word.lemma_lower()] += dict_noun_word_to_counter[abbreviation]
            dict_noun_word_to_counter.pop(abbreviation, None)
            dict_noun_lemma_to_examples.pop(abbreviation, None)
            example_lst_abbreviation_new = []
            for example in example_lst_abbreviation:
                np = example[0].replace(abbreviation, elaboration)
                score = example[1] + len(dict_elaboration_to_lst[elaboration]) - 1
                np_lemmas_lst = set()
                for lemma_word in example[2]:
                    if lemma_word == abbreviation:
                        for token in dict_elaboration_to_lst[elaboration]:
                            np_lemmas_lst.add(token)
                        continue
                    np_lemmas_lst.add(lemma_word)
                example_lst_abbreviation_new.append(np, score, np_lemmas_lst)
            dict_noun_lemma_to_examples[word.lemma_lower()].extend(example_lst_abbreviation_new)


def extend_ignore_words(ignore_words):
    new_ignore_words = set()
    for word in ignore_words:
        word_as_tokens = parse_medical_data.nlp(word)
        for token in word_as_tokens:
            lemma_word = token.lemma_.lower()
            new_ignore_words.add(lemma_word)
    ignore_words.extend(new_ignore_words)


def convert_examples_to_clustered_data(examples, ignore_words, host_and_port, has_umls_server):
    utils_clustering.dict_span_to_topic_entry = {}
    utils_clustering.dict_span_to_rank = {}
    utils_clustering.dict_noun_lemma_to_counter = {}
    utils_clustering.dict_noun_lemma_to_examples = {}
    utils_clustering.dict_noun_lemma_to_noun_words = {}
    utils_clustering.dict_noun_word_to_counter = {}
    utils_clustering.dict_noun_lemma_to_span = {}
    dict_noun_lemma_to_synonyms = {}
    dict_full_np_to_sentences = {}
    span_lst = set()
    dict_span_to_counter = {}
    dict_longest_span_to_counter = {}
    dict_word_to_lemma = {}
    counter = 0
    dict_sentence_to_span_lst = {}
    valid_span_lst = set()
    extend_ignore_words(ignore_words)
    head_lst = set()
    for biggest_noun_phrase, head_span, all_valid_nps_lst, sentence in examples:
        head_lst.add(head_span.lemma_.lower())
        span = valid_expansion_utils.get_tokens_as_span(biggest_noun_phrase)
        dict_sentence_to_span_lst[sentence] = dict_sentence_to_span_lst.get(sentence, [])
        if span in span_lst:
            counter = utils_clustering.update_recurrent_span(dict_sentence_to_span_lst, sentence, span,
                                                             dict_longest_span_to_counter, all_valid_nps_lst,
                                                             dict_span_to_counter, counter, dict_full_np_to_sentences)
            continue
        dict_sentence_to_span_lst[sentence].append(span)
        span_lst.add(span)
        tokens_already_counted = set()
        lemma_already_counted = set()
        is_valid_example = False
        for word in biggest_noun_phrase:
            if word in tokens_already_counted:
                continue
            lemma_word = word.lemma_.lower()
            dict_word_to_lemma[word.text.lower()] = lemma_word
            is_valid_example |= utils_clustering.add_word_collection_to_data_structures(word, tokens_already_counted,
                                                                                        lemma_already_counted,
                                                                                        all_valid_nps_lst, span,
                                                                                        ignore_words)
        if is_valid_example:
            dict_full_np_to_sentences[span] = [sentence]
            counter = utils_clustering.update_new_valid_example(span, dict_longest_span_to_counter, all_valid_nps_lst,
                                                                dict_span_to_counter, valid_expansion_utils, counter,
                                                                valid_span_lst)
    utils_clustering.create_synonym_dicts(dict_noun_lemma_to_synonyms, host_and_port, has_umls_server)
    dict_lemma_to_synonyms = utils_clustering.create_dicts_for_words_similarity(dict_word_to_lemma, host_and_port,
                                                                                has_umls_server)
    dict_lemma_to_synonyms.update(dict_noun_lemma_to_synonyms)
    topics_dict = {k: v for k, v in
                   sorted(utils_clustering.dict_noun_lemma_to_examples.items(), key=lambda item: len(item[1]),
                          reverse=True) if set(dict_noun_lemma_to_synonyms[k]).intersection(head_lst)}

    return topics_dict, dict_span_to_counter, dict_word_to_lemma, dict_lemma_to_synonyms, \
           dict_longest_span_to_counter, dict_noun_lemma_to_synonyms, utils_clustering.dict_noun_lemma_to_noun_words, \
           utils_clustering.dict_noun_lemma_to_counter, utils_clustering.dict_noun_word_to_counter, \
           dict_full_np_to_sentences



import heapq
from itertools import combinations
import torch
from hierarchybuilder.DAG import DAG_utils as DAG_utils

cos = torch.nn.CosineSimilarity(dim=0, eps=1e-08)


def dfs_update_marginal_gain(visited, node, dist_matrix, k, dist=1):  # function for dfs
    if node not in visited:
        visited.append(node)
        for neighbour in node.children:
            if dist < dist_matrix.get(hash(k) - hash(neighbour), float('inf')):
                dist_matrix[hash(k) - hash(neighbour)] = dist
            dfs_update_marginal_gain(visited, neighbour, dist_matrix, k, dist + 1)


def dfs(visited, node):  # function for dfs
    if node not in visited:
        visited.append(node)
        for neighbour in node.children:
            dfs(visited, neighbour)


def calculate_dist_from_set_to_vertex(S, v, dist_matrix):
    min_dist = float('inf')
    for u in S:
        dist_u_v = dist_matrix.get(hash(str(hash(u))) - hash(str(hash(v))), float('inf'))
        if dist_u_v < min_dist:
            min_dist = dist_u_v
    return min_dist


def get_rep_from_group(S, y, dist_matrix, global_index_to_similar_longest_np, topic_lst):
    dist = calculate_dist_from_set_to_vertex(S, y, dist_matrix)
    return get_rep(y, dist, global_index_to_similar_longest_np, topic_lst)


def get_rep(y, dist, global_index_to_similar_longest_np, x, topic_lst):
    rep = get_value_by_cosineSimilarity_format(global_index_to_similar_longest_np, y, x, topic_lst)
    return rep


def calculate_marginal_gain(x, dist_matrix, S_rep, k, dict_object_to_desc,
                            global_index_to_similar_longest_np, topic_lst):
    marginal_val = 0
    S_rep_new = {}
    for y in dict_object_to_desc[hash(x)]:
        if dist_matrix.get(hash(k) - hash(y), 1) == 0:
            continue
        else:
            marginal_val_y = S_rep.get(hash(y), 0)
            if x == y:
                gain_x = get_value_by_cosineSimilarity_format(global_index_to_similar_longest_np, x, x, topic_lst)
                S_rep_new[hash(y)] = gain_x
                marginal_val += (gain_x - marginal_val_y)
            else:
                dist = dist_matrix[hash(str(hash(x))) - hash(str(hash(y)))]
                S_rep_new[hash(y)] = get_rep(y, dist, global_index_to_similar_longest_np, x, topic_lst)
                marginal_val += (S_rep_new[hash(y)] - marginal_val_y)
    return marginal_val, S_rep_new


def get_value_by_cosineSimilarity_format(global_index_to_similar_longest_np, current_node, main_node, topic_lst):
    if current_node in topic_lst:
        label_lst = DAG_utils.get_labels_of_children(current_node.children)
        label_lst_minus_children_labels = current_node.label_lst - label_lst
        marginal_gain = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                                label_lst_minus_children_labels)
        return marginal_gain
    try:
        cos_similarity_val = cos(current_node.weighted_average_vector, main_node.weighted_average_vector)
    except:
        # print("Error in weighted_average_vector")
        # print(current_node.span_lst)
        # print(current_node.list_of_span_as_lemmas_lst)
        cos_similarity_val = 0.5
    frequency = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                        current_node.label_lst)
    marginal_gain = frequency * (cos_similarity_val ** 2)
    return marginal_gain
    # label_lst = DAG_utils.get_labels_of_children(current_node.children)
    # label_lst_minus_children_labels = current_node.label_lst - label_lst
    # node_frequency = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
    #                                                         label_lst_minus_children_labels)
    # try:
    #     cos_similarity_val = cos(current_node.weighted_average_vector, main_node.weighted_average_vector)
    # except:
    #     print("Error in weighted_average_vector")
    #     print(current_node.span_lst)
    #     print(current_node.list_of_span_as_lemmas_lst)
    #     cos_similarity_val = 0.5
    # marginal_gain = node_frequency * (cos_similarity_val ** 2)
    # return marginal_gain


def get_value_in_score_format(dist, global_index_to_similar_longest_np, node):
    label_lst = DAG_utils.get_labels_of_children(node.children)
    label_lst_minus_children_labels = node.label_lst - label_lst
    # labels_children = label_lst - label_lst_minus_children_labels
    leaves_labels = DAG_utils.get_labels_of_leaves(node.children)
    marginal_gain = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                            label_lst_minus_children_labels) ** 2
    marginal_gain += DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                             leaves_labels) * (node.score - 1)
    marginal_gain += DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                             label_lst - leaves_labels) * (node.score - 1)
    marginal_gain = marginal_gain / (dist + 1)
    return marginal_gain


def get_value_in_pow_format(dist, global_index_to_similar_longest_np, label_lst_node, children):
    label_lst = DAG_utils.get_labels_of_children(children)
    label_lst = label_lst_node - label_lst
    marginal_gain = (DAG_utils.get_frequency_from_labels_lst
                     (global_index_to_similar_longest_np, label_lst) ** 2) / (dist + 1)
    return marginal_gain


def compute_value_for_each_node(x, dist_matrix, dict_object_to_desc, dict_node_to_rep,
                                global_index_to_similar_longest_np, topic_lst):
    Q = [x]
    visited = [x]
    dist_matrix[hash(x)] = 0
    dict_object_to_desc[hash(x)] = []
    rep_matrix = {}
    counter = 0
    total_gain = 0
    while Q:
        v = Q.pop()
        dict_object_to_desc[hash(x)].append(v)
        if x == v:
            x_v = hash(x)
        else:
            x_v = hash(str(hash(x))) - hash(str(hash(v)))
        dist_matrix[x_v] = dist_matrix.get(x_v, 0)
        for u in v.children:
            if u not in visited:
                x_u = hash(str(hash(x))) - hash(str(hash(u)))
                dist_matrix[x_u] = dist_matrix[x_v] + 1
                x_u_marginal_gain = get_value_by_cosineSimilarity_format(global_index_to_similar_longest_np, u, x,
                                                                         topic_lst)
                rep_matrix[x_u] = x_u_marginal_gain
                total_gain += x_u_marginal_gain
                counter += 1
                Q.append(u)
                visited.append(u)
    total_gain += get_value_by_cosineSimilarity_format(global_index_to_similar_longest_np, x, x, topic_lst)
    rep_matrix[hash(x)] = total_gain
    dict_node_to_rep[list(x.span_lst)[0]] = rep_matrix
    return total_gain


def get_all_group_with_intersection_greater_than_X(selected_np_objects, threshold_intersection=0.7):
    objects_set_more_than_threshold_intersection = []
    for comboSize in range(2, len(selected_np_objects)):
        for combo in combinations(range(len(selected_np_objects)), comboSize):
            intersection = selected_np_objects[combo[0]].label_lst
            intersection_set = [selected_np_objects[combo[0]]]
            max_object_idx = max(combo[1:], key=lambda idx: len(selected_np_objects[idx].label_lst))
            max_labels_val = max(len(selected_np_objects[max_object_idx].label_lst), len(intersection))
            for i in combo[1:]:
                intersection = intersection & selected_np_objects[i].label_lst
                intersection_set.append(selected_np_objects[i])
            if len(intersection) > threshold_intersection * max_labels_val:
                objects_set_more_than_threshold_intersection.append(intersection_set)
    return objects_set_more_than_threshold_intersection


common_span_remove_lst = []


def remove_unselected_np_objects(parent_np_object, selected_np_objects):
    # list of unselected children to remove from parent
    remove_lst = []
    for child in parent_np_object.children:
        if child not in selected_np_objects:
            remove_lst.append(child)
            if parent_np_object in child.parents:
                child.parents.remove(parent_np_object)
    for node in remove_lst:
        if node.frequency > 1:
            common_span_remove_lst.append(node)
    for np_object in remove_lst:
        parent_np_object.children.remove(np_object)


def set_cover(children, np_object_parent, global_index_to_similar_longest_np):
    covered = set()
    selected_np_objects = []
    counted_labels = set()
    while True:
        # np_object = max(children, key=lambda np_object: len(
        #     np_object_parent.label_lst.intersection(np_object.label_lst - covered)), default=None)
        try:
            np_object = max(children, key=lambda np_object: DAG_utils.get_frequency_from_labels_lst(
                global_index_to_similar_longest_np, np_object.label_lst - covered), default=None)
        except:
            # for child in children:
            #     print(child.span_lst)
            #     raise Exception("There is error in the max function")
            continue
        if not np_object:
            break
        if DAG_utils.get_frequency_from_labels_lst(
                global_index_to_similar_longest_np, np_object.label_lst - covered) > 1:
            counted_labels.update(np_object_parent.label_lst.intersection(np_object.label_lst))
            selected_np_objects.append(np_object)
            covered.update(np_object.label_lst)
        else:
            break
    return selected_np_objects, counted_labels


def add_longest_nps_to_np_object_children(topic_object, labels, global_dict_label_to_object):
    longest_nps_lst = set()
    for label in labels:
        try:
            longest_nps_lst.add(global_dict_label_to_object[label])
        except:
            continue
    topic_object.add_children(longest_nps_lst)
    for longest_np in longest_nps_lst:
        if longest_np == topic_object:
            continue
        longest_np.parents.add(topic_object)


def sort_DAG_by_frequency(topic_object_lst, visited=[]):
    for topic_object in topic_object_lst:
        if topic_object in visited:
            continue
        visited.append(topic_object)
        topic_object.children = sorted(topic_object.children, key=lambda item: item.frequency, reverse=True)
        sort_DAG_by_frequency(topic_object.children, visited)


def DAG_contraction_by_set_cover_algorithm(topic_object_lst, global_dict_label_to_object,
                                           global_index_to_similar_longest_np):
    topic_object_lst = sorted(topic_object_lst, key=lambda item: item.frequency, reverse=True)
    visited_nodes = set()
    for topic_object in topic_object_lst:
        build_tree_from_DAG(topic_object, global_dict_label_to_object, visited_nodes,
                            global_index_to_similar_longest_np, set())
        # DAG_utils.check_symmetric_relation_in_DAG(topic_object_lst)
    sort_DAG_by_frequency(topic_object_lst)


def get_labels_from_visited_children(children, visited_nodes):
    visited_labels = set()
    for child in children:
        if child in visited_nodes:
            visited_labels.update(child.label_lst)
    return visited_labels


def build_tree_from_DAG(np_object, global_dict_label_to_object, visited_nodes,
                        global_index_to_similar_longest_np, visited_labels):
    if np_object in visited_nodes:
        return
    visited_nodes.add(np_object)
    if not np_object.children:
        return
    labels_covered_by_children = DAG_utils.get_labels_of_children(np_object.children)
    labels_covered_by_parent = np_object.label_lst - labels_covered_by_children
    # visited_labels.update(labels_covered_by_parent)
    # visited_labels.update(get_labels_from_visited_children(np_object.children, visited_nodes))
    # visited_labels = set()
    # visited_labels.update(labels_covered_by_parent)
    all_labels = np_object.label_lst - labels_covered_by_parent
    # unvisited_nodes = set(np_object.children) - visited_nodes
    selected_np_objects, counted_labels = set_cover(set(np_object.children), np_object,
                                                    global_index_to_similar_longest_np)
    unselected_nodes = set(np_object.children) - set(selected_np_objects)
    np_object.children = selected_np_objects.copy()
    np_object.children.extend(list(unselected_nodes))
    remove_unselected_np_objects(np_object, selected_np_objects)
    uncounted_labels = all_labels - counted_labels
    visited_labels.update(uncounted_labels)
    add_longest_nps_to_np_object_children(np_object, uncounted_labels, global_dict_label_to_object)
    if selected_np_objects:
        for np_object_child in selected_np_objects:
            build_tree_from_DAG(np_object_child, global_dict_label_to_object, visited_nodes,
                                global_index_to_similar_longest_np, set())


def is_ancestor_in_S(v, S, visited):
    visited.add(v)
    for parent in v.parents:
        if parent in visited:
            continue
        if parent in S:
            return True
        is_ancestor = is_ancestor_in_S(parent, S, visited)
        if is_ancestor:
            return True
    return False


def get_all_nodes(node, visited):
    for child in node.children:
        if child in visited:
            continue
        visited.add(child)
        get_all_nodes(child, visited)


def extract_top_k_concept_nodes_greedy_algorithm(k, topic_lst, global_index_to_similar_longest_np):
    dist_matrix = {}
    dict_object_to_desc = {}
    dict_node_to_rep = {}
    all_object_np_lst = []
    for node in topic_lst:
        dfs(all_object_np_lst, node)
    all_labels = set()
    for node in all_object_np_lst:
        all_labels.update(node.label_lst)
        node.marginal_val = compute_value_for_each_node(node, dist_matrix, dict_object_to_desc, dict_node_to_rep,
                                                        global_index_to_similar_longest_np, topic_lst)
    S = []
    all_nodes = set()
    heap_data_structure = all_object_np_lst
    heapq.heapify(heap_data_structure)
    already_counted_labels = set()
    S_rep = {}
    counter = 0
    while heap_data_structure:
        x = heapq.heappop(heap_data_structure)
        if x in all_nodes:
            continue
        # is_ancestor_already_in_S = is_ancestor_in_S(x, S, set())
        # if is_ancestor_already_in_S:
        #     continue
        marginal_val_x, S_rep_new = calculate_marginal_gain(x, dist_matrix, S_rep, k,
                                                            dict_object_to_desc,
                                                            global_index_to_similar_longest_np, topic_lst)
        if x.marginal_val > marginal_val_x + 0.1:
            x.marginal_val = marginal_val_x
            heapq.heappush(heap_data_structure, x)
            continue
        uncounted_labels = []
        for label in x.label_lst:
            if label not in already_counted_labels:
                uncounted_labels.append(label)
        # uncounted_labels_frequency = DAG_utils.get_frequency_from_labels_lst(
        #     global_index_to_similar_longest_np, uncounted_labels)
        # if uncounted_labels_frequency < 5:
        #     continue
        if len(uncounted_labels) < 0.3 * len(x.label_lst):
            continue
        already_counted_labels.update(x.label_lst)
        for key, value in S_rep_new.items():
            S_rep[key] = value
        S.append(x)
        get_all_nodes(x, all_nodes)
        dist_matrix[hash(k) - hash(x)] = 0
        counter += 1
        dfs_update_marginal_gain([], x, dist_matrix, k)
        if counter == k:
            break
    return S, already_counted_labels, all_labels



from quantulum3 import parser
from hierarchybuilder.combine_spans import utils as combine_spans_utils


def from_quantity_to_range(quantity_lst):
    quantity_lst.sort()
    if len(quantity_lst) == 1:
        return str(quantity_lst[0])
    max_val = max(quantity_lst)
    min_val = min(quantity_lst)
    range_number = str(min_val) + "-" + str(max_val)
    return range_number


def normalized_quantity_node(node):
    unit_to_quantity_dict = {}
    span_to_surface_dict = {}
    surface_to_unit = {}
    valid_span_lst = set()
    for span in node.span_lst:
        try:
            quants = parser.parse(span)
        except:
            continue
        for quant in quants:
            if quant.unit.name and quant.unit.name != 'dimensionless':
                valid_span_lst.add(span)
                span_to_surface_dict[span] = span_to_surface_dict.get(span, set())
                span_to_surface_dict[span].add(quant.surface)
                surface_to_unit[quant.surface] = quant.unit.name
                unit_to_quantity_dict[quant.unit.name] = unit_to_quantity_dict.get(quant.unit.name, set())
                unit_to_quantity_dict[quant.unit.name].add(quant.value)
    is_valid_normalized = False
    for explicit_, quantity in unit_to_quantity_dict.items():
        if len(quantity) > 0.5 * len(node.span_lst) and len(quantity) > 1:
            is_valid_normalized = True
            break
    if is_valid_normalized:
        most_frequent_span = combine_spans_utils.get_most_frequent_span(valid_span_lst)
        surface_lst = span_to_surface_dict[most_frequent_span]
        for surface in surface_lst:
            unit = surface_to_unit[surface]
            quantity_lst = unit_to_quantity_dict[unit]
            range_number = from_quantity_to_range(list(quantity_lst))
            normalized_quantity = range_number + " " + unit
            most_frequent_span = most_frequent_span.replace(surface, normalized_quantity)
    else:
        most_frequent_span = combine_spans_utils.get_most_frequent_span(node.span_lst)
    return most_frequent_span


def get_quantity_for_nodes(object_lst, visited=set()):
    for node in object_lst:
        if node in visited:
            continue
        normalized_quantity_node(node)
        visited.add(node)
        get_quantity_for_nodes(node.children, visited)



import uvicorn
from hierarchybuilder.UMLS import umls_loader
import json
from fastapi import FastAPI, APIRouter

router = APIRouter()
app = FastAPI()


@app.post("/create_synonyms_dictionary/")
def create_synonyms_dictionary(words: str):
    word_lst = json.loads(words)
    dict_lemma_to_synonyms = {}
    for word in word_lst:
        synonyms = umls_loader.umls_loader.get_term_aliases(word)
        dict_lemma_to_synonyms[word] = set()
        dict_lemma_to_synonyms[word].add(word)
        synonyms = set(synonyms)
        if synonyms:
            for synonym in synonyms:
                if synonym != word:
                    dict_lemma_to_synonyms[word].add(synonym)
    return {"synonyms": dict_lemma_to_synonyms}


@app.post("/get_broader_terms/")
def get_broader_terms(terms: str, relation_type: str):
    broader_terms = set()
    term_lst = json.loads(terms)
    for term in term_lst:
        broader_terms = umls_loader.umls_loader.get_broader_term(term, relation_type)
        if not broader_terms:
            continue
    return {"broader_terms": broader_terms}


def create_umls_server(host="127.0.0.1", port=5000, umls_relations_file_path='../UMLS_data/MRREL.RRF',
                       umls_synonymous_file_path='../UMLS_data/MRCONSO.RRF'):
    umls_loader.load_umls_data(umls_relations_file_path, umls_synonymous_file_path)
    uvicorn.run(app, host=host, port=port, log_level="info")



from nltk.corpus import wordnet
import requests
from hierarchybuilder.expansions import valid_expansion_utils
import json
from nltk.corpus import stopwords

# import umls_loader
stop_words = set(stopwords.words('english'))
tied_deps = ['compound', 'mwe', 'name', 'nummod']
neglect_deps = ['neg', 'case', 'mark', 'auxpass', 'aux', 'cop', 'nummod', 'quantmod', 'nmod:npmod', 'det:predet']

dict_span_to_topic_entry = {}
dict_span_to_rank = {}
dict_noun_lemma_to_counter = {}
dict_noun_lemma_to_examples = {}
dict_noun_lemma_to_noun_words = {}
dict_noun_word_to_counter = {}
dict_noun_lemma_to_span = {}


def filter_dict_by_lst(topic_lst):
    global dict_noun_lemma_to_examples, dict_noun_lemma_to_counter
    dict_noun_lemma_to_examples = {key: dict_noun_lemma_to_examples[key] for key in dict_noun_lemma_to_examples if
                                   key in topic_lst}
    dict_noun_lemma_to_counter = {key: dict_noun_lemma_to_counter[key] for key in dict_noun_lemma_to_counter if
                                  key in topic_lst}
    return dict_noun_lemma_to_examples, dict_noun_lemma_to_counter


def set_cover():
    covered = set()
    topic_lst = set()
    noun_to_spans_lst = []
    for noun, tuples_span_lst in dict_noun_lemma_to_examples.items():
        spans_lst = [tuple_span[0] for tuple_span in tuples_span_lst]
        noun_to_spans_lst.append((noun, set(spans_lst)))
    while True:
        item = max(noun_to_spans_lst, key=lambda s: len(s[1] - covered))
        if len(item[1] - covered) > 0:
            covered.update(item[1])
            topic_lst.add(item[0])
        else:
            break
    return topic_lst


def update_recurrent_span(dict_sentence_to_span_lst, sentence, span, dict_longest_span_to_counter, all_valid_nps_lst,
                          dict_span_to_counter, counter, dict_full_np_to_sentences):
    if span not in dict_sentence_to_span_lst[sentence]:
        dict_sentence_to_span_lst[sentence].append(span)
        if span in dict_longest_span_to_counter:
            dict_longest_span_to_counter[span] += 1
            dict_full_np_to_sentences[span].append(sentence)
            counter += 1
        for sub_span in all_valid_nps_lst:
            dict_span_to_counter[
                valid_expansion_utils.get_tokens_as_span(sub_span[0])] = dict_span_to_counter.get(
                valid_expansion_utils.get_tokens_as_span(sub_span[0]), 0) + 1
    return counter


def update_new_valid_example(span, dict_longest_span_to_counter, all_valid_nps_lst,
                             dict_span_to_counter, valid_expansion_utils, counter, valid_span_lst):
    dict_longest_span_to_counter[span] = 1
    # if not valid_span_lst:
    #     dict_span_to_counter[span] = dict_span_to_counter.get(span, 0) + 1
    #     print("There is longest expansion that isn't in the all_valid_nps_lst")
    for sub_span in all_valid_nps_lst:
        dict_span_to_counter[valid_expansion_utils.get_tokens_as_span(sub_span[0])] = dict_span_to_counter.get(
            valid_expansion_utils.get_tokens_as_span(sub_span[0]), 0) + 1
    valid_span_lst.add(span)
    counter += 1
    return counter


def initialize_token_expansions_information(all_valid_nps_lst, token, lemma_word, span):
    expansions_contain_word = []
    for sub_span in all_valid_nps_lst:
        if token in sub_span[0]:
            np_span = valid_expansion_utils.get_tokens_as_span(sub_span[0])
            dict_span_to_rank[np_span] = sub_span[1]
            lemma_lst = from_tokens_to_lemmas(sub_span[0])
            expansions_contain_word.append((np_span, sub_span[1], lemma_lst))
    # if lemma_word == 'null':
    #     print("something is strange")
    dict_noun_lemma_to_examples[lemma_word] = dict_noun_lemma_to_examples.get(lemma_word, [])
    dict_noun_lemma_to_examples[lemma_word].append((span, expansions_contain_word))
    dict_noun_lemma_to_counter[lemma_word] = dict_noun_lemma_to_counter.get(lemma_word, 0)
    dict_noun_lemma_to_counter[lemma_word] += 1
    dict_span_to_topic_entry[span] = dict_span_to_topic_entry.get(span, set())
    dict_span_to_topic_entry[span].add(lemma_word)


def add_word_collection_to_data_structures(word, tokens_already_counted, lemma_already_counted,
                                           all_valid_nps_lst, span, ignore_words):
    is_valid_example = False
    if word.pos_ in ['NOUN', 'ADJ', 'PROPN']:
        compound_noun = combine_tied_deps_recursively_and_combine_their_children(word)
        compound_noun.sort(key=lambda x: x.i)
        is_valid_example = False
        token = word
        # for token in compound_noun:
        if token.dep_ in ['quantmod', 'nummod'] or token.text == '-':
            return False
        lemma_token = token.lemma_.lower()
        if lemma_token in lemma_already_counted:
            return is_valid_example
        lemma_already_counted.add(lemma_token)
        if lemma_token in ignore_words:
            return is_valid_example
        if token not in tokens_already_counted:
            dict_noun_word_to_counter[lemma_token] = dict_noun_word_to_counter.get(lemma_token, 0) + 1
            dict_noun_lemma_to_noun_words[lemma_token] = dict_noun_lemma_to_noun_words.get(lemma_token,
                                                                                           set())
            tokens_already_counted.add(token)
            initialize_token_expansions_information(all_valid_nps_lst, token, lemma_token, span)
            is_valid_example = True
    return is_valid_example


def combine_dicts(dict_clustered_spans_last, dict_clustered_spans):
    new_dict_clustered_spans = {}
    for phrase_main, phrases_lst in dict_clustered_spans.items():
        new_dict_clustered_spans[phrase_main] = []
        for phrase in phrases_lst:
            new_dict_clustered_spans[phrase_main].extend(dict_clustered_spans_last[phrase])
    return new_dict_clustered_spans


def combine_tied_deps_recursively_and_combine_their_children(head):
    combined_tied_tokens = [head]
    for child in head.children:
        if child.dep_ in tied_deps:
            temp_tokens = combine_tied_deps_recursively_and_combine_their_children(child)
            combined_tied_tokens.extend(temp_tokens)
    return combined_tied_tokens


def get_words_as_span(span_lst):
    span = ""
    idx = 0
    for word in span_lst:
        if idx != 0 and word != ',':
            span += ' '
        span += word
        idx += 1
    return span


def from_tokens_to_lemmas(tokens):
    lemma_lst = []
    for token in tokens:
        if token.dep_ in neglect_deps or token.lemma_ in stop_words or token.text in \
                ['-', '/', '//', ',', '(', ')', '.', ' ']:
            continue
        lemma_lst.append(token.lemma_.lower())
    return lemma_lst


def get_dict_sorted_and_filtered(dict_noun_lemma_to_value, abbreviations_lst, dict_noun_lemma_to_counter, head_lst):
    dict_noun_lemma_to_value = {key: dict_noun_lemma_to_value[key] for key in dict_noun_lemma_to_value if
                                (key in head_lst or dict_noun_lemma_to_counter[key] > 1) and
                                ord('A') <= ord(key[:1]) <= ord('z') and len(key) > 1 and key not in abbreviations_lst}
    return dict_noun_lemma_to_value


def get_synonyms_by_word(word):
    synonyms = set()
    for syn in wordnet.synsets(word):
        for l in syn.lemmas():
            synonyms.add(l.name())
    return synonyms


def create_dict_from_wordnet_word_to_exist_synonyms(lemma_lst, dict_lemma_to_synonyms):
    for lemma in lemma_lst:
        synonyms = get_synonyms_by_word(lemma)
        synonyms = [synonym for synonym in synonyms if synonym in lemma_lst]
        synonyms.append(lemma)
        dict_lemma_to_synonyms[lemma] = set(synonyms)


def create_dicts_for_words_similarity(dict_word_to_lemma, host_and_port, has_umls_server):
    lemma_lst = set()
    for _, lemma in dict_word_to_lemma.items():
        lemma_lst.add(lemma)
    dict_lemma_to_synonyms = {}
    create_dict_from_wordnet_word_to_exist_synonyms(lemma_lst, dict_lemma_to_synonyms)
    word_lst = list(lemma_lst)
    if has_umls_server:
        update_dictionary_umls_synonyms(dict_lemma_to_synonyms, word_lst, host_and_port)
    dict_lemma_to_synonyms = {k: v for k, v in
                              sorted(dict_lemma_to_synonyms.items(), key=lambda item: len(item[1]),
                                     reverse=True)}
    return dict_lemma_to_synonyms


def remove_non_relevant_synonyms_from_dict(dict_word_to_synonyms):
    for noun, synonyms in dict_word_to_synonyms.items():
        remove_lst = set()
        for synonym in synonyms:
            if synonym not in dict_word_to_synonyms:
                remove_lst.add(synonym)
        for synonym in remove_lst:
            synonyms.remove(synonym)


def update_dictionary_umls_synonyms(dict_word_to_synonyms, word_lst, host_and_port):
    for i in range(0, len(word_lst), 100):
        chunk = word_lst[i: i + 100]
        if 'null' in chunk:
            chunk.remove('null')
        try:
            post_data = json.dumps(chunk)
            dict_response = requests.post(host_and_port + '/create_synonyms_dictionary/',
                                          params={"words": post_data})
            output = dict_response.json()["synonyms"]
            dict_word_to_synonyms.update(output)
        except:
            for word in chunk:
                try:
                    post_data = json.dumps([word])
                    dict_response = requests.post(host_and_port + '/create_synonyms_dictionary/',
                                                  params={"words": post_data})
                    output = dict_response.json()["synonyms"]
                    dict_word_to_synonyms.update(output)
                except:
                    continue
    remove_non_relevant_synonyms_from_dict(dict_word_to_synonyms)


def create_synonym_dicts(dict_noun_lemma_to_synonyms, host_and_port, has_umls_server):
    global dict_noun_lemma_to_examples, dict_noun_lemma_to_counter
    dict_noun_lemma_to_examples_new = {}
    dict_noun_lemma_to_counter_new = {}
    dict_noun_lemma_to_examples = {k: v for k, v in
                                   sorted(dict_noun_lemma_to_examples.items(), key=lambda item: len(item[1]),
                                          reverse=True)}
    word_lst = list(dict_noun_lemma_to_examples.keys())
    if has_umls_server:
        update_dictionary_umls_synonyms(dict_noun_lemma_to_synonyms, word_lst, host_and_port)
    else:
        create_dict_from_wordnet_word_to_exist_synonyms(word_lst, dict_noun_lemma_to_synonyms)
        remove_non_relevant_synonyms_from_dict(dict_noun_lemma_to_synonyms)
    for word, synonyms in dict_noun_lemma_to_synonyms.items():
        dict_noun_lemma_to_examples_new[word] = []
        dict_noun_lemma_to_examples_new[word].extend(dict_noun_lemma_to_examples[word])
        dict_noun_lemma_to_counter_new[word] = dict_noun_lemma_to_counter[word]
        for synonym in synonyms:
            if word == synonym:
                continue
            for spans in dict_noun_lemma_to_examples[synonym]:
                dict_noun_lemma_to_examples_new[word].append((spans[0], spans[1]))
            dict_noun_lemma_to_counter_new[word] += dict_noun_lemma_to_counter[synonym]
    dict_noun_lemma_to_counter_new = {k: v for k, v in
                                      sorted(dict_noun_lemma_to_counter_new.items(), key=lambda item: item[1])}
    dict_noun_lemma_to_span_new = {k: v for k, v in
                                   sorted(dict_noun_lemma_to_examples_new.items(), key=lambda item: len(item[1]),
                                          reverse=True)}
    dict_noun_lemma_to_examples = dict_noun_lemma_to_span_new
    dict_noun_lemma_to_counter = dict_noun_lemma_to_counter_new



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hierarchybuilder",
    version="0.0.6",
    author="Itay Yair and Hillel Taub-Tabib and Yoav Goldberg",
    author_email="iy24592@gmail.com",
    description="implementation of hierarchy builder",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    url="https://github.com/itayair//hierarchybuilder",
    install_requires=[
        'fastapi==0.94.1',
        'nltk==3.8.1',
        'pandas==1.5.3',
        'quantulum3==0.8.1',
        'requests==2.28.2',
        'scikit_learn==1.2.0',
        'setuptools==60.2.0',
        'spacy==3.0.9',
        'torch==1.13.1',
        'tqdm==4.64.1',
        'transformers==4.26.1',
        'uvicorn==0.21.0'
    ],
    project_urls={
        "Bug Tracker": "https://github.com/itayair/hierarchybuilder/-/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.7"
)
